/* Machine or compiler-dependent portions of kernel
 * Turbo-C version for PC
 * Copyright 1991 Phil Karn, KA9Q
 */
#include "global.h"
#include "commands.h"
#ifdef MSDOS
#include <dos.h>
#include "pc.h"
#endif
#include "proc.h"
#include "socket.h"

#if !defined(_lint)
static char rcsid[] OPTIONAL = "$Id: ksubr.c,v 1.17 1996/09/04 01:34:13 root Exp root $";
#endif

#if !defined(TNOS_68K) && !defined(UNIX)
static char *Taskers[] = {
	"",
	"DoubleDos",
	"DesqView",
	"Windows",
        "DOS 5/6",
        "OS/2 DOS",
        "DPMI"
};

#endif
#ifdef TNOS_68K
/* Template for contents of jmp_buf for 68302 */
struct env {
	unsigned	_pc;
	unsigned	_d0;
	unsigned	_d1;
	unsigned	_d2;
	unsigned	_d3;
	unsigned	_d4;
	unsigned	_d5;
	unsigned	_d6;
	unsigned	_d7;
	unsigned	_a0;
	unsigned	_a1;
	unsigned	_a2;
	unsigned	_a3;
	unsigned	_a4;
	unsigned	_a5;
	unsigned	_a6;
	unsigned	_a7;
	unsigned	_xtra;
};

#endif

#ifdef USE_SETSTACK
/*
 * This needs to be extern because it is referenced by setstack()
 */
void *newstackptr;
void setstack(void);
#endif

#ifdef UNIX
/*
 * There are several different ways to implement jmp_buf's.  We use macros to
 * extract the fields.  Note that these must expand into lvalues, because we
 * load them during process initialization.  Current code only uses _SP and
 * _PC; this may change.
 *
 * I should fold the DOS code into this, but then I need to worry about
 * segments.  No thanks.
 */

#ifdef M_UNIX
/*
 * I am unsure of these; the i386 contents are not documented in 3.2.2.  I
 * guessed at them by comparing "info regs" with the contents of a jmp_buf
 * immediately after a setjmp().
 */
#define _PC(p) (p->env[5])
#define _SP(p) (p->env[4])
#define _BP(p) (p->env[3])
#endif

#ifdef linux
#define _PC(p) (p->env->__pc)
#define _SP(p) (p->env->__sp)
#define _BP(p) (p->env->__bp)
#endif

#ifdef sun
#define _PC(p) (p->env[3])
#define _SP(p) (p->env[2])
#endif

#endif

int stkutil (struct proc *pp);

#if !defined(UNIX) && !defined(__DJGPP__)
static unsigned short oldNull;
static int chkintstk (void);


int
dostackcheck(argc,argv,p)
int argc;
char *argv[];
void *p OPTIONAL;
{
    return setbool(&Stkchk,"Stack checking code",argc,argv);
}
#endif

static void pproc(struct proc *pp);



void
kinit()
{
#if !defined(UNIX) && !defined(TNOS_68K) && !defined(__DJGPP__)
	int i;

	/* Initialize interrupt stack for high-water-mark checking */
	for(i=0;i<512;i++)
		Intstk[i] = STACKPAT;
#endif

#if !defined(UNIX) && !defined(__DJGPP__)
	/* Remember location 0 pattern to detect null pointer derefs */
	oldNull = *(unsigned short *)NULL;
#endif
#ifdef MSDOS
	/* Initialize signal queue */
	Ksig.wp = Ksig.rp = Ksig.entry;
#endif
}
/* Print process table info
 * Since things can change while ps is running, the ready proceses are
 * displayed last. This is because an interrupt can make a process ready,
 * but a ready process won't spontaneously become unready. Therefore a
 * process that changes during ps may show up twice, but this is better
 * than not having it showing up at all.
 */
int
ps(argc,argv,p)
int argc OPTIONAL;
char *argv[] OPTIONAL;
void *p OPTIONAL;
{
register struct proc *pp;
int i;

#ifdef TNOS_68K
	tprintf("Uptime %s Stack %x\n",tformat(secclock()),stacksiz());
#else
	tprintf("Uptime %s\n",tformat(secclock()));
#endif

#if 0
	tprintf("ksigs %lu queued %lu hiwat %u woken %lu nops %lu dups %u\n",Ksig.ksigs,
	 Ksig.ksigsqueued,Ksig.maxentries,Ksig.ksigwakes,Ksig.ksignops,Ksig.duksigs);
	Ksig.maxentries = 0;
	tprintf("kwaits %lu nops %lu from int %lu\n",
	 Ksig.kwaits,Ksig.kwaitnops,Ksig.kwaitints);
#endif

#ifdef USE_SETSTACK
	tputs("PID      stksize  maxstk   event    fl  in  out  name\n");
#else
#ifdef MSDOS
	tputs("PID      SP       stksize  event    fl  in  out  name\n");
#else
	tputs("PID      SP       stksize  maxstk   event    fl  in  out  name\n");
#endif
#endif

	for (pp = Susptab;pp != NULLPROC;pp = pp->next)
		pproc(pp);

	for (i = 0; i < PHASH; i++)
		for (pp = Waittab[i]; pp != NULLPROC; pp = pp->next)
			pproc(pp);

	for (pp = Rdytab; pp != NULLPROC; pp = pp->next)
		pproc(pp);

	if(Curproc != NULL)
		pproc(Curproc);

	return 0;
}

static void
pproc(pp)
struct proc *pp;
{
#if defined(USE_SETSTACK) || defined(MSDOS)
	tprintf("%8.8lx %8.8lx %8.8lx %8.8lx %c%c%c %3d %3d  %s\n",
#else
	tprintf("%8.8lx %8.8lx %8.8lx %8.8lx %8.8lx %c%c%c %3d %3d  %s\n",
#endif
		FP_SEG(pp),
#ifndef USE_SETSTACK
#ifdef MSDOS
		pp->env[0].__esp,
#else
		FP_SEG(_SP(pp)),
#endif
#endif
		(int32) pp->stksize,
#ifndef MSDOS
		(int32) stkutil(pp),
#endif
		FP_SEG(pp->event),
#ifdef __DJGPP__
		pp->flags.istate ? 'I' : ' ',
		pp->flags.waiting ? 'W' : ' ',
		pp->flags.suspend ? 'S' : ' ',
#else
		pp->i_state ? 'I' : ' ',
		(pp->state & WAITING) ? 'W' : ' ',
		(pp->state & SUSPEND) ? 'S' : ' ',
#endif
		pp->input,pp->output,pp->name);
}

int
stkutil(pp)
struct proc *pp;
{
	unsigned i;
	register int16 *sp;

#ifdef UNIX
#ifdef linux
	if (FP_SEG(pp->stack) == 0xC0000000LU)
#else
	if (FP_SEG(pp->stack) == 0x80000000LU)
#endif
		return ((int)pp->stksize);
#endif
	i = pp->stksize;
	for(sp = pp->stack;*sp == STACKPAT && sp < pp->stack + pp->stksize;sp++)
		i--;
	return ((int)i);
}
/* Return number of used words in interrupt stack. Note hardwired value
 * for stack size; this is also found in the various .asm files
 */
#if !defined(TNOS_68K) && !defined(UNIX) && !defined(__DJGPP__)
static int
chkintstk()
{
	register int i;
	register int16 *cp;

	for(i=512,cp = Intstk; i != 0 && *cp == STACKPAT; cp++)
		i--;
	return i;
}
#endif

/* Verify that stack pointer for current process is within legal limits;
 * also check that no one has dereferenced a null pointer
 */
#if 0	/* was def UNIX */
static void
#ifdef PROTOTYPES
__chkstk_internal(int16 spp)
#else
__chkstk_internal(spp)
    int16 spp;
#endif
{
	int16 *sp, *sbase, *stop;

	sp = &spp;		/* close enough for government work */
	sbase = Curproc->stack;
#ifdef UNIX
#ifdef linux
        if (sbase == (int16 *) 0xC0000000LU)
#else
        if (sbase == (int16 *) 0x80000000LU)
#endif
#else
 	if (sbase == (int16 *) 0)
#endif
		return;	/* Main task -- too hard to check */
	stop = sbase + Curproc->stksize;
	if(sp < sbase || sp >= stop){
		tprintf("Stack violation, process %s\n",Curproc->name);
		log (-1, "Stack violation, process %s\n",Curproc->name);
		tprintf("SP = %lx, legal stack range [%lx,%lx)\n",
			ptol(sp),ptol(sbase),ptol(stop));
		log (-1, "SP = %lx, legal stack range [%lx,%lx)\n",
			ptol(sp),ptol(sbase),ptol(stop));
		tflush();
		rflush();
		killself();
	}
}	
#endif


#if !defined(UNIX) && !defined(__DJGPP__)
void
chkstk()
{
#ifdef UNIX
	__chkstk_internal(0);	/* must have an argument to take address of */
#else
	int16 *sbase;
	int16 *stop;
	int16 *sp;
#ifdef MULTITASK
	extern int Nokeys;		/* indicates we are shelled out [pc.c]*/
#endif

#ifndef TNOS_68K
	sp = MK_FP(_SS,_SP);
	if(_SS == _DS){
		/* Probably in interrupt context */
		return;
	}
#else
	sp = getstack();
#endif
	sbase = Curproc->stack;
	if(sbase == NULL)
		return;	/* Main task -- too hard to check */

	stop = sbase + Curproc->stksize;
	if(sp < sbase || sp >= stop){
		printf("Stack violation, process %s\n",Curproc->name);
		printf("SP = %lx, legal stack range [%lx,%lx)\n",
		ptol(sp),ptol(sbase),ptol(stop));
		fflush(stdout);
		killself();
	}
	if(*(unsigned short *)NULL != oldNull){
#ifdef MULTITASK
        if(!Nokeys)     /* don't complain if we are shelled out */
#endif
			printf("WARNING: Location 0 smashed, process %s\n",Curproc->name);
        *(unsigned short *)NULL = oldNull;
		fflush(stdout);
	}
#endif
}

#endif


/* Machine-dependent initialization of a task */
void
psetup(pp,iarg,parg1,parg2,pc)
struct proc *pp;	/* Pointer to task structure */
int iarg OPTIONAL;	/* Generic integer arg */
void *parg1 OPTIONAL;	/* Generic pointer arg #1 */
void *parg2 OPTIONAL;	/* Generic pointer arg #2 */
void (*pc)(int,void*,void*);           /* Initial execution address */
{
#ifdef MSDOS
static int32 *stktop;
#else
#ifndef USE_SETSTACK
static int16 *stktop; /* KF8NH was "int", this breaks 32-bit code */
#endif
#endif

#if 0	/* was ndef UNIX */
struct env *ep;
#endif

#ifdef USE_SETSTACK
	/*
	 * setstack() is a routine borrowed from WAMPES.  It's actually
	 * an assembly language routine inside setsp.c.  TNOS needs to
	 * return the proc pointer, so I have to play a little game here.
	 * I do a setjmp() call, establishing this piece of code as the
	 * process context.  Note that we won't really be on the new
	 * process stack when setjmp() returns the first time.  The
	 * actual switchover doesn't happen until the process gets scheduled
	 * and the setjmp() returns for the second time.
	 */
	if (setjmp(pp->env))	{
		newstackptr = Curproc->stack + Curproc->stksize;
		setstack();
		/* We are now running on the new process stack.  We can call
		 * the function and everything will be hunky-dory. */
		(Curproc->func)(Curproc->iarg, Curproc->parg1, Curproc->parg2);

		/* Clean up, just in case the function returns. */
		killself();
	} else {
		/* Still running in the old context.  Stuff the function
		 * pointer away so we can call the function later. */
		pp->func = pc;
		/* Run with interrupts on */
		pp->i_state = 1;
		return;
	}
#else

	/* Set up stack to make it appear as if the user's function was called
	 * by killself() with the specified arguments. When the user returns,
	 * killself() automatically cleans up.
	 *
	 * First, push args on stack in reverse order, simulating what C
	 * does just before it calls a function.
	 */
#ifdef MSDOS
	stktop = (int32 *)pp->stack + pp->stksize;
	*--stktop = (int32)parg2;
	*--stktop = (int32)parg1;
	*--stktop = (int32)iarg;

	/* Now push the entry address of killself(), simulating the call to
	 * the user function.
	 */
	*--stktop = (int32)killself;
#else
	stktop = (int16 *)(pp->stack + pp->stksize); /* KF8NH as above */

#ifdef UNIX
#ifdef __GNUC__
#define THEMASK 0xFFFF0000LU
#else
#define THEMASK 0xFFFF0000
#endif

#ifdef IS_LITTLE_ENDIAN
	*--stktop = (((unsigned long) parg2 & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) parg2 & 0x0000FFFF);
	*--stktop = (((unsigned long) parg1 & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) parg1 & 0x0000FFFF);
	*--stktop = (((unsigned long) iarg & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) iarg & 0x0000FFFF);
	*--stktop = (((unsigned long) killself & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) killself & 0x0000FFFF);
#else
	*--stktop = ((unsigned long) parg2 & 0x0000FFFF);
	*--stktop = (((unsigned long) parg2 & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) parg1 & 0x0000FFFF);
	*--stktop = (((unsigned long) parg1 & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) iarg & 0x0000FFFF);
	*--stktop = (((unsigned long) iarg & THEMASK) >> 16) & 0x0000FFFF;
	*--stktop = ((unsigned long) killself & 0x0000FFFF);
	*--stktop = (((unsigned long) killself & THEMASK) >> 16) & 0x0000FFFF;
#endif
#else
#ifdef	LARGEDATA
	*--stktop = FP_SEG(parg2);
#endif
	*--stktop = FP_OFF(parg2);
#ifndef TNOS_68K
#ifdef	LARGEDATA
	*--stktop = FP_SEG(parg1);
#endif
	*--stktop = FP_OFF(parg1);
	*--stktop = iarg;
		
	/* Now push the entry address of killself(), simulating the call to
	 * the user function.
	 */
#ifdef	LARGECODE
	*--stktop = FP_SEG(killself);
#endif
#endif
	*--stktop = FP_OFF(killself);
#endif
#endif

	/* Set up task environment. Note that for Turbo-C, the setjmp
	 * sets the interrupt enable flag in the environment so that
	 * interrupts will be enabled when the task runs for the first time.
	 * Note that this requires newproc() to be called with interrupts
	 * enabled!
	 */
	setjmp(pp->env);
#ifdef UNIX
#ifdef sun
	_SP(pp) = (int)stktop;
	_PC(pp) = (int)pc;
#else
	_SP(pp) = stktop;
	_BP(pp) = stktop;
	_PC(pp) = pc;
#endif
#else
#ifndef TNOS_68K
	pp->env[0].__esp = (int32)stktop;
	pp->env[0].__eip = (int32)pc;
	pp->env[0].__ebp = 0;		/* Anchor stack traces */
	/* Task initially runs with interrupts on */
	pp->flags.istate = 1;
#else
	ep = (struct env *)pp->env;
	ep->_d1 = parg1;
	ep->_d0 = iarg;
	ep->_a7 = stktop;
	ep->_pc = pc;
#endif
#endif
	/* Task initially runs with interrupts on */
	pp->i_state = 1;
#endif
}
unsigned
#ifdef PROTOTYPES
phash(volatile void *event)
#else
phash(event)
volatile void *event;
#endif
{
#ifdef __DJGPP__
	/* If PHASH is a power of two, this will simply mask off the
	 * higher order bits
	 */
	return (int)event % PHASH;
#else
	register unsigned x;

	/* Fold the two halves of the pointer */
#ifdef UNIX
	x = (unsigned) event;
#else
#ifndef TNOS_68K
	x = FP_SEG(event) ^ FP_OFF(event);
#else
	x = hiword(((long)event)) ^ loword(((long)event));
#endif
#endif

	/* If PHASH is a power of two, this will simply mask off the
	 * higher order bits
	 */
	return x % PHASH;
#endif
}
